import sys
from pathlib import Path

_MODULES_DIR = Path.cwd() / "modules"

for zipfile in _MODULES_DIR.glob("**/*.zip"):
    sys.path.append(zipfile.as_posix())

_JOBS_DIR = Path.cwd() / "jobs"
sys.path.append(_JOBS_DIR.as_posix())



import zipfile
from pathlib import Path
from typing import Optional
from superglue.environment.config import MODULES_PATH
from superglue.core.types import SuperglueModuleType
from superglue.core.components.base import SuperglueComponent
from superglue.core.components.tests import SuperglueTests
from superglue.environment.variables import SUPERGLUE_S3_BUCKET, SUPERGLUE_IAM_ROLE


class SuperglueModule(SuperglueComponent):
    def __init__(self, module_name: str, tests: Optional[SuperglueTests] = None, *args, **kwargs):
        self.tests = tests or SuperglueTests()

        super(SuperglueModule, self).__init__(
            *args,
            root_dir=MODULES_PATH,
            component_name=module_name,
            component_type="superglue_module",
            bucket=SUPERGLUE_S3_BUCKET,
            iam_role=SUPERGLUE_IAM_ROLE,
            **kwargs,
        )

    @property
    def module_name(self) -> str:
        return self.component_name

    @property
    def module_root_path(self) -> Path:
        return self.root_dir / self.module_name

    @property
    def module_inner_path(self) -> Path:
        return self.module_root_path / self.module_name

    @property
    def zipfile(self) -> Path:
        return self.module_root_path / f"{self.module_name}.zip"

    @property
    def s3_zipfile_path(self) -> str:
        relative_path = self.zipfile.relative_to(self.module_root_path)
        return f"{self.s3_path}/{relative_path}"

    @property
    def module_test_path(self) -> Path:
        return self.tests.modules_test_dir / self.module_name

    @property
    def module_tests_file(self):
        return self.module_test_path / f"test_{self.module_name}.py"

    @classmethod
    def new(cls, module_name: str) -> SuperglueModuleType:
        return cls(module_name)

    @classmethod
    def get(cls, module_name: str) -> SuperglueModuleType:
        sg_module = cls(module_name)
        if not sg_module.module_root_path.exists():
            raise FileNotFoundError(f"No superglue module {module_name} exists.")
        return sg_module

    @classmethod
    def from_version(cls, module_name: str, version_number: int) -> SuperglueModuleType:
        sg_module = cls(module_name=module_name, version_number=version_number)
        if not sg_module.module_root_path.exists():
            raise FileNotFoundError(f"No superglue module {module_name} exists.")
        return sg_module

    def save(self) -> None:
        if not self.module_root_path.exists():
            self.module_inner_path.mkdir(parents=True, exist_ok=True)

            init_py = self.module_inner_path / "__init__.py"
            init_py.touch(exist_ok=True)
            self.save_version_file()
            self.create_zip()
            self.save_tests()
            print(f"created new superglue module {self.module_name}")

        else:
            print(f"shared python module {self.module_name} already exists.")

    def deploy(self, force: Optional[bool] = False) -> None:
        if force:
            print(f"Forcing deployment of superglue module {self.module_name}")
            self.sync()
        elif self.is_deployable:
            self.sync()
            print(f"Superglue module {self.module_name} successfully deployed!")
        elif self.is_edited:
            print(f"Superglue module {self.module_name} has edits in progress. Please run superglue package")
        else:
            print(f"Superglue module {self.module_name} up to date in S3. Nothing to deploy.")

    def delete(self) -> None:
        raise NotImplementedError

    def create_zip(self) -> None:
        with zipfile.ZipFile(self.zipfile, mode="w") as zip_file:
            for file in self.module_root_path.glob("**/*.py"):
                content = file.read_text()
                content = content.encode("utf-8")
                rel_path = file.relative_to(self.module_root_path).as_posix()
                zip_file.writestr(rel_path, content)

    def package(self, force: Optional[bool] = False) -> None:

        if force:
            print(f"Forcing packaging of superglue module {self.module_name}")
            self.create_zip()
            self.save_version_file()
        elif self.is_edited:
            self.create_zip()
            self.save_version_file()
            print(f"Superglue module {self.module_name} has been successfully packaged!")
        else:
            print(f"Superglue module {self.module_name} package is up to date!")

    def save_tests(self) -> None:
        if not self.module_test_path.exists():
            jinja = self.get_jinja_environment()
            tests_template = jinja.get_template("module_test.template.py.txt")
            test_content = tests_template.render(module=self.module_name)

            self.module_test_path.mkdir(exist_ok=True, parents=True)
            self.module_tests_file.touch(exist_ok=True)
            self.module_tests_file.write_text(test_content)
            print(f"Tests created for superglue module {self.module_name}")
        else:
            print(f"Tests already exists for superglue module {self.module_name}")



from typing import TypeVar

SuperglueJobType = TypeVar("SuperglueJobType", bound="SuperglueJob")
SuperglueModuleType = TypeVar("SuperglueModuleType", bound="SuperglueModule")
SuperglueComponentType = TypeVar("SuperglueComponentType", bound="SuperglueComponent")
SuperglueCommandType = TypeVar("SuperglueCommandType", bound="SuperglueCommand")
SuperglueMakefileType = TypeVar("SuperglueMakefileType", bound="SuperglueMakefile")
SuperglueTestsType = TypeVar("SuperglueTestsType", bound="SuperglueTests")
SuperglueFilesType = TypeVar("SuperglueFilesType", bound="SuperglueFiles")



from pathlib import Path

JOBS_PATH = Path.cwd() / "jobs"
MODULES_PATH = Path.cwd() / "modules"
NOTEBOOKS_PATH = Path.cwd() / "notebooks"
TOOLS_PATH = Path.cwd() / "tools"
TESTS_PATH = Path.cwd() / "tests"



import boto3
import inspect
from typing import List, Any, Callable
from types import ModuleType
from superglue.cli.command import SuperglueCommand
from superglue.core.types import SuperglueCommandType
from superglue.environment.variables import SUPERGLUE_AWS_ACCOUNT


def get_commands(module: ModuleType) -> List[SuperglueCommandType]:
    commands = []
    for _, _class in inspect.getmembers(module, inspect.isclass):
        if _class != SuperglueCommand:
            if issubclass(_class, SuperglueCommand):
                commands.append(_class)
    return commands


def validate_account(func: Callable) -> Callable:
    def wrapper(*args, **kwargs) -> Any:
        sts = boto3.client("sts")
        account_id = int(sts.get_caller_identity()["Account"])

        if account_id != SUPERGLUE_AWS_ACCOUNT:
            print(f"superglue expects account {SUPERGLUE_AWS_ACCOUNT} but account is {account_id}")
            exit(1)
        print(f"\n -- Using AWS Account {SUPERGLUE_AWS_ACCOUNT} --")

        return func(*args, **kwargs)

    return wrapper


def get_parser_name(name: str) -> str:
    return name.split(".")[-1]


def yes_no_confirmation(msg: str) -> None:

    while True:

        selection = input(f"{msg} : Are you sure? : [y/n] -> ").lower()

        if selection not in ["y", "n"]:
            print("please make a valid selection.")
        elif selection == "y":
            break
        elif selection == "n":
            exit(0)
        else:
            print("Invalid selection.")
            exit(1)



from superglue.cli.command import SuperglueCommand
from superglue.cli.utils import validate_account, get_parser_name


parser_name = get_parser_name(__name__)
cli_help = "--> Interact with superglue jobs"


class BaseArgs:

    args = {("-n", "--name"): {"required": True, "help": "Name of the superglue job (directory)"}}


class New(BaseArgs, SuperglueCommand):

    help = "--> Create a new superglue job directory."

    def __call__(self):
        job = self.project.job.new(self.cli_args.name)
        job.save()


class Status(SuperglueCommand):

    help = "--> Print the current status of all superglue jobs"

    @validate_account
    def __call__(self):
        table = self.project.get_pretty_table()

        for job in self.project.jobs:
            table.add_row(job.pretty_table_row)

        print(table)


class Package(SuperglueCommand):

    help = "--> Packages superglue jobs and dependent modules which have changed since the last issued package command"

    def __call__(self) -> None:

        if self.project.jobs.edited():
            for job in self.project.jobs.edited():
                for module in job.modules():
                    module.package()
                job.package()
        else:
            print("No changes found to superglue jobs. Nothing to package. All jobs are up to date.")


class Build(BaseArgs, SuperglueCommand):

    help = "--> Create or overwrite the 'deployment.yml' file for a given job."

    def __call__(self) -> None:
        job = self.project.job.get(self.cli_args.name)
        job.render()
        job.save_deployment_config()


class Deploy(BaseArgs, SuperglueCommand):

    help = "--> Manually deploy a superglue job. Jobs must first be up to date using the package command."

    def __call__(self) -> None:
        job = self.project.job.get(self.cli_args.name)

        for module in job.modules():
            module.package()
            module.deploy()

        job.deploy()



from superglue.cli.command import SuperglueCommand
from superglue.cli.utils import get_parser_name, validate_account


parser_name = get_parser_name(__name__)
cli_help = "--> Interact with superglue modules"


class BaseArgs:

    args = {("-n", "--name"): {"required": True, "help": "Name of the superglue module (directory)"}}


class New(BaseArgs, SuperglueCommand):

    help = "--> Create a new superglue module to share across superglue jobs."

    def __call__(self) -> None:
        module = self.project.module.new(self.cli_args.name)
        module.save()


class Status(SuperglueCommand):

    help = "--> Print the status of all superglue modules."

    @validate_account
    def __call__(self) -> None:
        table = self.project.get_pretty_table()

        for module in self.project.modules:
            table.add_row(module.pretty_table_row)

        print(table)


class Package(SuperglueCommand):

    help = "--> Packages superglue modules which have changed since the last issued package command"
    args = {
        ("-f", "--force"): {
            "action": "store_true",
            "default": False,
            "help": "Force repackaging of all superglue modules, regardless of the local or S3 state.",
        }
    }

    def __call__(self) -> None:

        if self.cli_args.force:
            for module in self.project.modules:
                module.package(self.cli_args.force)

        elif self.project.modules.edited():
            for module in self.project.modules.edited():
                module.package()

        else:
            print("Superglue modules are up to date. Nothing to package!")


class Deploy(SuperglueCommand):

    help = "--> Manually deploy a superglue module. Modules must first be up to date using the package command."
    args = {
        ("-n", "--name"): {"required": True, "help": "Name of the superglue module (directory)"},
        ("-f", "--force"): {
            "action": "store_true",
            "default": False,
            "help": "Force repackaging of all superglue modules, regardless of the local or S3 state.",
        },
    }

    @validate_account
    def __call__(self) -> None:
        module = self.project.module.get(self.cli_args.name)
        module.deploy(self.cli_args.force)



import boto3
import json
import botocore
from io import BytesIO
from pathlib import Path
from hashlib import md5
from abc import ABC, abstractmethod
from multiprocessing import Pool, cpu_count
from jinja2 import Environment, PackageLoader
from typing import Optional, Dict, Tuple, List
from superglue.core.types import SuperglueComponentType
from superglue.environment.variables import SUPERGLUE_IAM_ROLE, SUPERGLUE_S3_BUCKET


class SuperglueComponent(ABC):
    def __init__(
        self,
        root_dir: Path,
        component_name: str,
        component_type: str,
        version_number: Optional[int] = None,
        bucket: Optional[str] = SUPERGLUE_S3_BUCKET,
        iam_role: Optional[str] = SUPERGLUE_IAM_ROLE,
    ) -> None:

        self.root_dir = root_dir
        self.component_name = component_name
        self.component_type = component_type
        self.bucket = bucket
        self.iam_role = iam_role
        self._version_number = version_number

    def __eq__(self, other: SuperglueComponentType) -> bool:
        try:
            return self.component_name == other.component_name
        except AttributeError:
            return False

    @property
    def component_path(self) -> Path:
        return self.root_dir / self.component_name

    @property
    def version_file(self) -> Path:
        return self.component_path / ".version"

    @property
    def version(self) -> Dict:
        if not self.version_file.exists():
            self.save_version_file(increment_version=False)
        return json.load(self.version_file.open())

    @property
    def version_number(self) -> int:
        if self._version_number:
            return self._version_number
        return self.version.get("version_number", 0)

    @property
    def next_version_number(self) -> int:
        return self.version_number + 1

    @property
    def s3_path(self) -> str:
        return f"s3://{self.bucket}/superglue/{self.component_type}/{self.component_name}/version={self.version_number}"

    @property
    def s3_prefix(self) -> str:
        return f"superglue/{self.component_type}/{self.component_name}/version={self.version_number}"

    @property
    def s3_version_path(self) -> str:
        return f"{self.s3_prefix}/{self.component_name}/.version"

    @property
    def is_edited(self) -> bool:
        return self.version != self.get_version_hashes()

    @property
    def is_deployable(self) -> bool:
        return self.is_edited is False and self.version != self.fetch_s3_version()

    @property
    def status(self) -> Tuple[str, str]:

        if self.is_edited:
            return "edits in progress", "out of sync"

        elif self.is_deployable:
            return "up to date", "out of sync"

        else:
            return "up to date", "in sync"

    @property
    def pretty_table_row(self) -> List[str]:
        return [self.component_name, self.component_type, *self.status, self.version_number]

    @staticmethod
    def get_jinja_environment() -> Environment:
        """
        used to get a jinja2 templating environment.
        Returns: Jinja2 Environment
        """
        loader = PackageLoader(package_name="superglue", package_path="templates")
        return Environment(loader=loader, trim_blocks=True, lstrip_blocks=True)

    def component_files(self) -> List[Path]:
        file_list = []
        filters = [".DS_Store", ".empty"]

        for path in self.component_path.glob("**/*"):
            if path.is_file() and path.name not in filters:
                file_list.append(path)
        return file_list

    def get_relative_path(self, path: Path) -> str:
        return path.relative_to(self.root_dir).as_posix()

    def hash_file(self, path: Path) -> Tuple[str, str]:

        relative_path = self.get_relative_path(path)
        md5_hash = md5()

        with path.open("rb") as data:
            for chunk in iter(lambda: data.read(4096), b""):
                md5_hash.update(chunk)
            return relative_path, md5_hash.hexdigest()

    def get_version_hashes(self) -> Dict[str, str]:
        version_hashes = {}
        for path in self.component_files():
            if path.name != ".version":
                key, digest = self.hash_file(path)
                version_hashes[key] = digest
        return version_hashes

    def save_version_file(self, increment_version: Optional[bool] = True) -> None:
        version_hashes = self.get_version_hashes()

        if increment_version:
            version_hashes["version_number"] = self.next_version_number
        else:
            version_hashes["version_number"] = 0

        json.dump(version_hashes, self.version_file.open(mode="w"), indent=4)

    def upload_object_to_s3(self, path: Path) -> None:
        s3_client = boto3.client("s3")
        relative_path = self.get_relative_path(path)
        print(f"Uploading -- s3://{self.bucket}/{self.s3_prefix}/{relative_path}")
        s3_client.upload_file(path.as_posix(), self.bucket, f"{self.s3_prefix}/{relative_path}")

    def sync(self) -> None:
        with Pool(cpu_count()) as pool:
            files = self.component_files()
            pool.map(self.upload_object_to_s3, files)

    def fetch_s3_version(self) -> Dict[str, str]:
        s3_client = boto3.client("s3")
        try:
            with BytesIO() as buffer:
                s3_client.download_fileobj(self.bucket, self.s3_version_path, buffer)
                buffer.seek(0)
                return json.load(buffer)
        except botocore.exceptions.ClientError as e:
            code = e.args[0].partition(":")[2].strip()
            if code == "Not Found":
                return {}
            raise e

    @abstractmethod
    def deploy(self) -> None:
        pass

    @abstractmethod
    def delete(self) -> None:
        pass

    @abstractmethod
    def save(self) -> None:
        pass



from superglue.cli.utils import validate_account
from superglue.cli.command import SuperglueCommand


__version__ = "0.3.2"


class Version(SuperglueCommand):

    args = {}
    help = "--> Print the current version of superglue and exit."

    def __call__(self) -> None:
        print(f"The AWS Glue Deployment Utility -- superglue version :: {__version__}")


class Account(SuperglueCommand):

    help = "--> Print the AWS account number that superglue is configured to use."

    @validate_account
    def __call__(self) -> None:
        pass


class Init(SuperglueCommand):

    help = "--> Creates a new superglue project structure."

    def __call__(self) -> None:
        self.project.create()
        print("superglue project initialized!")


class Package(SuperglueCommand):

    help = "--> Packages all superglue jobs and modules which have been edited since the last package was issued."

    def __call__(self) -> None:
        edited = self.project.modules.edited()
        if edited:
            for module in edited:
                module.package()
        else:
            print("No changes in superglue modules found. Nothing to package.")

        edited = self.project.jobs.edited()

        if edited:
            for job in edited:
                job.package()
        else:
            print("No changes in superglue jobs found. Nothing to package.")


class Status(SuperglueCommand):

    help = "--> Prints the current status of all superglue jobs and modules."

    @validate_account
    def __call__(self) -> None:
        table = self.project.get_pretty_table()

        for module in self.project.modules:
            table.add_row(module.pretty_table_row)

        for job in self.project.jobs:
            table.add_row(job.pretty_table_row)

        print(table)


class Deploy(SuperglueCommand):

    help = "--> Deploys all superglue jobs and modules which have been changed since the last issued deployment."

    args = {("-p", "--package"): {"action": "store_true", "default": False}}

    def __call__(self) -> None:
        if self.cli_args.package:
            print("Packaging before deployment")
            for module in self.project.modules.edited():
                module.package()

            for job in self.project.jobs.edited():
                job.package()
        else:
            if self.project.modules.edited() or self.project.jobs.edited():
                print("Active edits in progress. Please run superglue package before attempting deployment")
                exit(1)

        deployable_modules = self.project.modules.deployable()
        if deployable_modules:
            for module in deployable_modules:
                module.deploy()
        else:
            print("All superglue modules are up to date in S3. Nothing to deploy.")

        deployable_jobs = self.project.jobs.deployable()

        if deployable_jobs:
            for job in deployable_jobs:
                job.deploy()
        else:
            print("All superglue jobs are up to date in S3. Nothing to deploy.")


class Generate(SuperglueCommand):

    help = "--> Used to generate superglue tests, and utility files"

    args = {
        ("-t", "--tests"): {"action": "store_true", "default": False},
        ("-m", "--makefile"): {"action": "store_true", "default": False},
    }

    def __call__(self) -> None:

        if self.cli_args.tests:
            for job in self.project.jobs:
                job.save_tests()

            for module in self.project.modules:
                module.save_tests()

        if self.cli_args.makefile:
            makefile = self.project.makefile.new()
            makefile.save()



from typing import List
from superglue.core.components.base import SuperglueComponentType


class SuperglueComponentList(list):
    def edited(self) -> List[SuperglueComponentType]:
        return [c for c in self if c.is_edited]

    def deployable(self) -> List[SuperglueComponentType]:
        return [c for c in self if c.is_deployable]



from argparse import ArgumentParser
from superglue.cli import root, job, module
from superglue.cli.utils import get_commands


def parse_args():

    # This is the root node of the main parser tree.
    parser = ArgumentParser()
    sub_parser = parser.add_subparsers(dest="command")
    sub_parser.required = True

    # set parser for all root level actions in the format
    # superglue "action" <options>
    for command in get_commands(root):
        p = sub_parser.add_parser(command.method(), help=command.help)
        for args, kwargs in command.args.items():
            p.add_argument(*args, **kwargs)
        p.set_defaults(command=command)

    # set subparser commands
    for commands in job, module:
        command_parser = sub_parser.add_parser(commands.parser_name, help=commands.cli_help)
        command_subparser = command_parser.add_subparsers()

        for command in get_commands(commands):
            p = command_subparser.add_parser(command.method(), help=command.help)
            for args, kwargs in command.args.items():
                p.add_argument(*args, **kwargs)
            p.set_defaults(command=command)

    # job_command_parser = sub_parser.add_parser("job")
    # job_command_subparser = job_command_parser.add_subparsers()
    # for command in job.commands():
    #     p = job_command_subparser.add_parser(command.method)
    #     for key, value in command.args.items():
    #         p.add_argument(*key, **value)
    #     p.set_defaults(command=command, method=command.method)

    # for job_command in JobCommands.commands():
    #     p = job_command_subparser.add_parser(job_command)
    #     p.add_argument("-n", "--name")
    #     p.set_defaults(command=JobCommands, method=job_command)

    # module_command_parser = sub_parser.add_parser("module")
    # module_command_subparser = module_command_parser.add_subparsers()
    #
    # for module_command in ModuleCommands.commands():
    #     p = module_command_subparser.add_parser(module_command)
    #     p.add_argument("-n", "--name")
    #     p.set_defaults(command=ModuleCommands, method=module_command)

    return parser.parse_args()
